#include <vector>
double TransfoX(double X, vector<double> &MinX, vector<double> &MaxX,
                vector<double> &ScalingBin, vector<double> &NewMinX,
                vector<double> &NewMaxX, vector<double> &Scale, int BinPerMM);

void GetBin(vector<double> &NBin, TH1F *&Hist, int StartBin, int EndBin);

void CalculateNewBin(vector<double> &MinX, vector<double> &MaxX,
                     vector<double> &NewMinX, vector<double> &NewMaxX,
                     vector<double> &BinWidth, vector<double> &ScalingBin,
                     vector<double> &Scale, vector<double> &NBin, int BinPerMM,
                     int mm);

void LinkBinsWithLines(vector<TF1 *> &lineFunctions, TH1F *&h);

void Linearisation4(int RunStart, int RunStop) {
  /*
    This function will Linearize the reconstructed position in X and Y for each
    FPMW of VAMOS.

    It will parse through an histogram mm per mm and store in a vector
    a scale to apply to each event depending on their position on the
    FPMW . Then, event per event it'll apply this scale which is made to ensure
    a linear distribution of the event in the FPMW.

    Input :
      int RunStart : Index of the starting run. The run to be treated should be
      named linea_{index}.root

      int RunStop : Index of the final run.

      Root file of data

    Output :
      A root file similar to the input one with added branch with linearised
      position.
  */
  //========================================================================
  //-------------------Open tree and define Hist----------------------------
  //========================================================================

  // Import the chain
  TChain *ch = new TChain("PhysicsTree");
  for (int run = RunStart; run <= RunStop; run++) {
    ch->Add(Form("../../root/analysis/linea_%d.root", run));
  } // End of run loop

  // Declaration of the histograms storing the reconstructed position
  // We define 2 histograms storing the current position and 2 holding
  // Previously non analysed position loaded from a file.

  std::vector<TH1F *> PositionX, PositionY, OldPosX(4), OldPosY(4);
  for (int i = 1; i <= 4; i++) {
    std::string PosXName = "hPositionXDet" + std::to_string(i);
    std::string PosXTitle = "PositionXDet" + std::to_string(i);
    std::string PosYName = "hPositionYDet" + std::to_string(i);
    std::string PosYTitle = "PositionYDet" + std::to_string(i);
    TH1F *PosX;
    TH1F *PosY;
    if (i == 1) {
      PosX = new TH1F(PosXName.c_str(), PosXTitle.c_str(), 1000, 0, 40);
      PosY = new TH1F(PosYName.c_str(), PosYTitle.c_str(), 1000, 0, 60);
    } // End if det 1

    if (i == 2) {
      PosX = new TH1F(PosXName.c_str(), PosXTitle.c_str(), 1000, 0, 63);
      PosY = new TH1F(PosYName.c_str(), PosYTitle.c_str(), 1000, 0, 92);

    } // End if det 2

    if (i == 3 || i == 4) {
      PosX = new TH1F(PosXName.c_str(), PosXTitle.c_str(), 1000, -1000, 0);
      PosY = new TH1F(PosYName.c_str(), PosYTitle.c_str(), 1000, 0, 160);
    } // End if det 3 or 4

    PositionX.push_back(PosX);
    PositionY.push_back(PosY);
  } // End of i loop
  TH1F *Det1 = new TH1F("Det1", "Det1", 1000, 0, 40);
  TH1F *Det1Adjusted = new TH1F("Det1Adjusted", "Det1Adjusted", 1000, 0, 40);
  TH1F *VarBin;

  //========================================================================
  //-----------------------Importing conditional var------------------------
  //========================================================================
  TFPMWPhysics *FPMW = new TFPMWPhysics();
  ch->SetBranchAddress("FPMW", &FPMW);

  // Loading the OldPosX and OldPosY histogram
  TFile *InputHist =
      TFile::Open("BackupHist.root"); // If it doesnt work import backup
  if (!InputHist || InputHist->IsZombie()) {
    cerr << "Error opening file !" << endl;
    return;
  }
  for (int i = 1; i <= 4; i++) {
    std::string PosXName = "hPositionXDet" + std::to_string(i);
    std::string PosXTitle = "PositionXDet" + std::to_string(i);
    std::string PosYName = "hPositionYDet" + std::to_string(i);
    std::string PosYTitle = "PositionYDet" + std::to_string(i);

    OldPosX[i - 1] = (TH1F *)InputHist->Get(PosXName.c_str());
    OldPosY[i - 1] = (TH1F *)InputHist->Get(PosYName.c_str());

    OldPosX[i - 1]->SetDirectory(nullptr);
    OldPosY[i - 1]->SetDirectory(nullptr);

  } // End of i loop

  InputHist->Close();
  //========================================================================
  //---------------------Calculate New Bin----------------------------------
  //========================================================================

  bool Verbose = true; // Just for some cout

  // Declaration of all vectors necessary to hold the information
  // Outputted by the function for each part of the detector.
  vector<vector<double>> MinXAll, MaxXAll, NewMinXAll, NewMaxXAll,
      NewBinWidthAll, ScalingBinAll, ScaleAll, NBinAll;

  // This hold the Bin edges to make a control final histogram with variable
  // bin width.
  vector<double> VarBinHolder;

  // Defining some useful parameter for later.
  // For now it only works for det 0.
  int NumberBins = OldPosX[0]->GetNbinsX();
  double WidthHist = OldPosX[0]->GetBinLowEdge(NumberBins) +
                     OldPosX[0]->GetBinWidth(NumberBins) -
                     OldPosX[0]->GetBinLowEdge(1);
  int CurrentBin = 1;
  int BinPerMM = NumberBins / WidthHist;

  // Generate a function for each BinPerMM (each mm then) to transform the
  // distribution Into a non uniform distribution

  vector<TF1 *> LinePerMM;
  TH1F *HistBigBin = (TH1F *)OldPosX[0]->Rebin(WidthHist, "HistBigBin");
  LinkBinsWithLines(LinePerMM, HistBigBin);

  // We will now parse through all the histogram, mm per mm and define a new
  // binning and scales that will linearise this histogram.
  for (int mm = OldPosX[0]->GetBinLowEdge(1);
       mm < OldPosX[0]->GetBinLowEdge(NumberBins); mm++) {

    // We define vector to hold the old and new binning, the bin width of each
    // bin and the scaling to apply to the bin width and to the X as well as the
    // number of event per bin.
    size_t N = BinPerMM;
    vector<double> MinX, MaxX, NewMinX(N + 1), NewMaxX(N), NewBinWidth,
        ScalingBin, Scale, NBin;
    int DetNbr = 0; // At some point we would need to iterate over all det

    // We fill Nbin with the value stored in each bin of the current range.
    GetBin(NBin, OldPosX[DetNbr], CurrentBin, CurrentBin + BinPerMM);
    // Fill The vector of edges of bin and bin width and scaling.
    CalculateNewBin(MinX, MaxX, NewMinX, NewMaxX, NewBinWidth, ScalingBin,
                    Scale, NBin, BinPerMM, mm);

    // Remember that there is another vector of vector holding all those values
    // .
    MinXAll.push_back(MinX);
    MaxXAll.push_back(MaxX);

    if (OldPosX[0]->GetBinContent(CurrentBin) == 0) {
      // In the case where the bin is empty we have a problem.
      // NewMinX and NewMax x will be undefined.
      // This check should be performed inside the CaculateNewBIn
      NewMinXAll.push_back(MinX);
      NewMaxXAll.push_back(MaxX);
      NewBinWidthAll.push_back(NewBinWidth);
      ScalingBinAll.push_back(ScalingBin);
      ScaleAll.push_back(Scale);
      NBinAll.push_back(NBin);
      for (int i = 0; i < N; i++) {
        VarBinHolder.push_back(MinX[i]);
      }
    }

    if (OldPosX[0]->GetBinContent(CurrentBin) != 0) {
      NewMinXAll.push_back(NewMinX);
      NewMaxXAll.push_back(NewMaxX);
      NewBinWidthAll.push_back(NewBinWidth);
      ScalingBinAll.push_back(ScalingBin);
      ScaleAll.push_back(Scale);
      NBinAll.push_back(NBin);

      for (int i = 0; i < N; i++) {
        // Filling a vector containing all the new bin edges.
        VarBinHolder.push_back(NewMinX[i]);
        if (Verbose == false) {
          cout << "MinX[i]   " << MinX[i];
          cout << "  NewMinX[i]  " << NewMinX[i] << endl;
          cout << " NBIN " << NBin[i] << endl;
        }                   // end verbose
      }                     // end loop on bin
    }                       // End if non empty
    CurrentBin += BinPerMM; // Changing Current Bin
  }                         // Loop on mm of the FPMW
  VarBinHolder.push_back(MaxXAll[MaxXAll.size() - 1][MaxXAll[3].size() - 1]);

  if (Verbose == false) {
    for (int i = 0; i < VarBinHolder.size(); i++) {
      cout << VarBinHolder[i] << " ";
    }
  }
  // Filling the VarBin Hist
  VarBin = new TH1F("VarBin", "VarBin", NumberBins, &VarBinHolder[0]);

  //========================================================================
  //---------------------------Treat Event----------------------------------
  //========================================================================

  int Nentries = ch->GetEntries();
  int GoodEntries = 0;

  for (int entries = 0; entries < Nentries; entries++) {
    ch->GetEntry(entries);
    if (entries % 100000 == 0) {
      cout << "\033[92m\r Processing tree..."
           << (double)entries / Nentries * 100 << "\% done" << flush;
    } // End cout loading

    // Verification that we are not on a empty event for the FPMW
    if (FPMW->DetectorNbr.size() > 0) {
      GoodEntries += 1;
      // Parsing through the different FPMW
      for (int DetectorNbr = 0; DetectorNbr < FPMW->DetectorNbr.size();
           DetectorNbr++) {
        double X = FPMW->PositionX[DetectorNbr];
        double Y = FPMW->PositionY[DetectorNbr];

        // Just filling two sanity check histo,
        //  To delete in final version
        if (DetectorNbr >= 0 && DetectorNbr < 4) {
          PositionX[DetectorNbr]->Fill(X);
          PositionY[DetectorNbr]->Fill(Y);
        }

        // For detector 0 -> In final version loop on detector must be done
        if (DetectorNbr == 0) {
          // For all bin in the MinXAll vector
          for (int bin = 0; bin < MinXAll.size(); bin++) {
            // If X is in the correct range
            if (X >= MinXAll[bin][0] &&
                X < MaxXAll[bin][MaxXAll[bin].size() - 1]) {
              // Apply the coorect transformation to X
              double XNew =
                  TransfoX(X, MinXAll[bin], MaxXAll[bin], ScalingBinAll[bin],
                           NewMinXAll[bin], NewMaxXAll[bin], ScaleAll[bin],
                           MinXAll[bin].size() - 1);

              // Fill of histogram for testing
              VarBin->Fill(XNew);
              Det1->Fill(X);
              Det1Adjusted->Fill(XNew);

              // Some cout
              if (Verbose == false) {
                cout << X << " " << XNew << "  "
                     << MaxXAll[bin][MaxXAll[bin].size() - 1] << endl;
                cout << "************************" << endl;
              }
            }
          } // end loop mm
        }   // End detNbr == 0
      }     // Loop detnbr
    }       // Non empty event

  } // End loop on entries

  cout << endl << "Good entries" << GoodEntries << endl;

  //========================================================================
  //---------------------Output histogram-----------------------------------
  //========================================================================

  TFile *OutputHist = new TFile("HistFPMW.root", "RECREATE");
  std::vector<TCanvas *> canvas;
  for (int i = 0; i < 4; i++) {
    std::string CanvasName = "c" + std::to_string(i);
    TCanvas *c = new TCanvas(CanvasName.c_str(), CanvasName.c_str(), 960, 540);
    c->Divide(2, 1);
    canvas.push_back(c);

    canvas[i]->cd(1);
    PositionX[i]->Draw("logY");
    PositionX[i]->Write();
    gPad->SetLogy();
    canvas[i]->cd(2);
    PositionY[i]->Draw("logY");
    PositionY[i]->Write();
    gPad->SetLogy();
  }
  TCanvas *CZoom = new TCanvas("CZoom", "CZoom", 1860, 540);
  CZoom->Divide(3, 1);
  CZoom->cd(1);
  Det1->Draw();
  CZoom->cd(2);
  Det1Adjusted->Draw("");
  Det1Adjusted->Write();
  CZoom->cd(3);
  VarBin->Draw("");
  VarBin->Write();
  gStyle->SetOptStat(11111111);

  OutputHist->Close();

} // end macro

//========================================================================
//--------------------------TransfoX--------------------------------------
//========================================================================

double TransfoX(double X, vector<double> &MinX, vector<double> &MaxX,
                vector<double> &ScalingBin, vector<double> &NewMinX,
                vector<double> &NewMaxX, vector<double> &Scale, int BinPerMM) {
  /*
     This take as input the original bin of the histogram and the new bin
     defined by the function CalculateNewBin And output the new calculated X

     This assume that the NewMinX and MinX have the same size !!!
     */
  double XNew = -1000;
  for (int i = 0; i < MinX.size(); i++) {
    if (X >= MinX[i] && X < MaxX[i]) {
      double Xold = X;
      double Scale2 = (X - MinX[i]) / (MaxX[i] - MinX[i]);
      XNew = Scale2 * (NewMaxX[i] - NewMinX[i]) + NewMinX[i];

      // Sanity check print
      if (XNew < NewMinX[i] || XNew > NewMaxX[i]) {
        cout << " FUNCTION  " << endl;
        cout << i << endl;
        cout << "MinX " << MinX[i] << " MaxX " << MaxX[i] << endl;
        cout << "Xold " << Xold << endl;
        cout << "NewMinX " << NewMinX[i] << " NewMaxX " << NewMaxX[i] << endl;
        cout << "Xnew  " << XNew << endl;
        cout << endl << endl;
      }
    }
  }
  return XNew;
} // end transfo x

//========================================================================
//--------------------------CalculateNewBin-------------------------------
//========================================================================

void CalculateNewBin(vector<double> &MinX, vector<double> &MaxX,
                     vector<double> &NewMinX, vector<double> &NewMaxX,
                     vector<double> &NewBinWidth, vector<double> &ScalingBin,
                     vector<double> &Scale, vector<double> &NBin, int BinPerMM,
                     int mm) {
  /*
    This function take as input the nimber of bin per mm and the starting
    position in the histogram and return several vector.

    Those vector old information on the edges of the bin of the initial
    histogram and the new calculated bin made for linearisation of the
    reconstructed position

    Most importantly it return SCALE which is used for event per event
    transformation of the reconstructed position in order to linearise the
    distribution for this specific range of value in this specific detector.

    Input :
      NBin : Number of event per bin in the range we are looking at.

      Bin Per MM : Number of bin in the range we are looking at.

      mm : Current position in the histogram

    Output :
      Min X , MaxX : Vector of the edges of the unlinearised bin.

      NewMinX, NewMaxX : Vector of the edges of the linearised bin.

      ScalingBin : Scaling applied to each bin in the range.

      Scale : Scaling to be applied to the Max edge of each bin. This is the
    main output of this macro

  */

  // Define some control variable
  double SumBin = 0, NTotBin, TotWidth = 0., NewTotWidth = 0.,
         IndexLastBin = BinPerMM;
  vector<double> BinWidth, NewBinWidthTheoretical;

  // Here we fill the vector for the unlinearised hist
  // Those can help to make cout to ensure that the histogram is nicely
  // imported.
  for (int i = 0; i < IndexLastBin; i++) {
    double d = i;
    double mmdouble = mm;
    MinX.push_back(mmdouble + d / (IndexLastBin));
    MaxX.push_back(mmdouble + (d + 1) / (IndexLastBin));
    SumBin += NBin[i];
    BinWidth.push_back(MaxX[i] - MinX[i]);
    TotWidth += BinWidth[i];
  }

  // NTotBin is here used as the optimal number of event per bin for it to be
  // linear. If we want the new distribution to not be linear in the range but
  // similar to a function We would need to change this value and define a
  // optimal number of event per BIN.
  NTotBin = SumBin / (IndexLastBin);
  NewMinX[0] = MinX[0]; // Just initialising the first new bin edge

  // Loop on the bin in the mm
  for (int i = 0; i < IndexLastBin; i++) {
    // We will multiply the bin width by the following value
    // by applying a transformation to the upper edge of the bin
    ScalingBin.push_back(NBin[i] / NTotBin);
    // The new bin should have a width equal to the following value
    NewBinWidthTheoretical.push_back(ScalingBin[i] * BinWidth[i]);

    if (i != 0) {
      // If we are not in the first bin the NewMinEdge should start at the
      // previous max edge
      NewMinX[i] = NewMaxX[i - 1];
    }
    // This would be the new max without the transformation of scaling bin. aka
    // if the bin were not linearised.
    double MaxXTemp = NewMinX[i] + BinWidth[i];
    // A scaling factor that will only be applied to the max edge of the bin
    // while giving a bin width given by the NewBinWidthTheoretical is to be
    // defined as follow
    Scale.push_back((NewMinX[i] + BinWidth[i] * ScalingBin[i]) / MaxXTemp);
    NewMaxX[i] = (NewMinX[i] + BinWidth[i]) * Scale[i];
    NewBinWidth.push_back(NewMaxX[i] - NewMinX[i]);
    NewTotWidth += NewBinWidth[i];

    if (true) {
      // Those are sanity check cout. To be turned on as needed.
      cout << " itération " << i << endl;
      cout << " MinX " << MinX[i] << " MaxX " << MaxX[i] << endl;
      cout << " NewMin " << NewMinX[i] << "  NewMax " << NewMaxX[i] << endl;
      cout << " Scaling  " << Scale[i] << " NewBinWidth " << NewBinWidth[i]
           << endl;
      cout << " Width of bin " << NewBinWidth[i] << "  expected width  "
           << NewBinWidthTheoretical[i] << endl;
      cout << " SumBin " << NTotBin  << endl;
      cout << endl << endl;
    }
  }
  // NewMinX is defined as all the edges of the bin not only the lower edges so
  // we need to assugb the last element manually.
  NewMinX[IndexLastBin + 1] = NewMaxX[IndexLastBin];
}

//========================================================================
//--------------------------GetBin----------------------------------------
//========================================================================
void GetBin(vector<double> &NBin, TH1F *&Hist, int StartBin, int EndBin) {
  /*
     Store into a input vector the value in a certain range of an histogram.
  */
  for (int bin = StartBin; bin < EndBin; bin++) {
    NBin.push_back(Hist->GetBinContent(bin));
  }
}

void LinkBinsWithLines(vector<TF1 *> &lineFunctions, TH1F *&h) {
  /*
     Take an histogram as input and output linear function
     linking the center of those histogram.
  */
  // Number of bins in the histogram
  int nBins = h->GetNbinsX();

  lineFunctions.clear();
  // Loop over the bins and create TF1 functions for each adjacent bin pair
  for (int i = 1; i < nBins; ++i) {
    // Get the centers of the current bin and the next bin
    double x1 = h->GetBinCenter(i);
    double y1 = h->GetBinContent(i);
    double x2 = h->GetBinCenter(i + 1);
    double y2 = h->GetBinContent(i + 1);

    // Calculate the slope and intercept of the line
    double slope = (y2 - y1) / (x2 - x1);
    double intercept = y1 - slope * x1;

    // Create the TF1 function for this line
    TF1 *line = new TF1(
        Form("line_%d", i),"[0]*x+[1]",
        x1, x2);
    line->SetParameter(0,slope);
    line->SetParameter(1,intercept);

    // Store the function in the vector
    lineFunctions.push_back(line);
  }
  if (true) {
    TCanvas *cline = new TCanvas("cline", "cline", 800, 600);
    h->Draw();
    for (size_t i = 0; i < lineFunctions.size(); ++i) {
      lineFunctions[i]->SetLineColor(kRed);
      lineFunctions[i]->Draw("same");
    } // end parser
  }   // end print
}
